"""
Contains the classes used for storing parameters and system states.
"""

import copy
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal
from matplotlib.colors import rgb2hex
from matplotlib.ticker import MaxNLocator

COLORMAP = "viridis"


def read_from_xyz(data_file: str, col: int = 4):
    """Read the input data from .xyz file."""
    with open(data_file, "r", encoding="utf-8") as file:
        tmp_list = [line.strip().split() for line in file]

    number_of_particles = int(tmp_list[0][0])
    number_of_frames = int(len(tmp_list) / (number_of_particles + 2))

    out_list = []
    line = 2
    for _ in range(number_of_frames):
        tmp = []
        for i in range(number_of_particles):
            tmp.append(tmp_list[line + i][col])
        out_list.append(tmp)
        line += number_of_particles + 2

    out_array = np.array(out_list, dtype=float).T
    return out_array


class StateUni:
    """
    Represents a state as a Gaussian.
    """

    def __init__(self, mean: float, sigma: float, area: float):
        self.mean = mean  # Mean of the Gaussian
        self.sigma = sigma  # Variance of the Gaussian
        self.area = area  # Area below the Gaussian
        self.peak = (
            area / sigma / np.sqrt(np.pi)
        )  # Height of the Gaussian peak
        self.perc = 0.0  # Fraction of data points classified in this state
        self.th_inf = [mean - 2.0 * sigma, -1]  # Lower thrashold of the state
        self.th_sup = [mean + 2.0 * sigma, -1]  # Upper thrashold of the state

    def build_boundaries(self, number_of_sigmas: float):
        """
        Sets the thresholds for the classification of data windows
        inside the state

        Args:
        - number of sigmas (float)
        """
        self.th_inf = [self.mean - number_of_sigmas * self.sigma, -1]
        self.th_sup = [self.mean + number_of_sigmas * self.sigma, -1]


class StateMulti:
    """
    Represents a state as a factorized Gaussian.
    """

    def __init__(self, mean: np.ndarray, sigma: np.ndarray, area: np.ndarray):
        self.mean = mean  # Mean of the Gaussians
        self.sigma = sigma  # Variance of the Gaussians
        self.area = area  # Area below the Gaussians
        self.perc = 0.0  # Fraction of data points classified in this state
        self.axis = 2.0 * sigma  # Axes of the state

    def build_boundaries(self, number_of_sigmas: float):
        """
        Sets the thresholds for the classification of data windows
        inside the state

        Args:
        - number of sigmas (float)
        """
        self.axis = number_of_sigmas * self.sigma  # Axes of the state


class UniData:
    """
    The input signals of the analysis.
    """

    def __init__(self, data_path: str):
        if data_path.endswith((".npz", ".npy", ".txt", ".xyz")):
            try:
                if data_path.endswith(".npz"):
                    with np.load(data_path) as data:
                        data_name = data.files[0]
                        self.matrix = np.array(data[data_name])
                elif data_path.endswith(".npy"):
                    self.matrix = np.load(data_path)
                elif data_path.endswith(".xyz"):
                    self.matrix = read_from_xyz(data_path)
                else:  # .txt file
                    self.matrix = np.loadtxt(data_path, dtype=float)
                print("\tOriginal data shape:", self.matrix.shape)
            except OSError as exc_msg:
                print(f"\tReading data from {data_path}: {exc_msg}")
                return
            except ValueError as exc_msg:
                print(f"\tReading data from {data_path}: {exc_msg}")
                return
        else:
            print("\tERROR: unsupported format for input file.")
            return

        self.num_of_particles = self.matrix.shape[0]
        self.num_of_steps = self.matrix.shape[1]
        self.range = np.array([np.min(self.matrix), np.max(self.matrix)])
        self.labels = np.array([])

    def print_info(self):
        """
        Prints information about the input data.
        """
        print("Number of particles:", self.num_of_particles)
        print("Number of steps:", self.num_of_steps)
        print("Data range:", self.range)

    def remove_delay(self, t_delay: int):
        """
        Removes a specified time delay from the data.

        Args:
        - t_delay (int): Number of steps to remove from the
        beginning of the data.
        """
        self.matrix = self.matrix[:, t_delay:]
        self.num_of_steps = self.matrix.shape[1]

    def smooth_lpf(self, sampling_freq: int, window: int):
        """
        Smooths the data using a digital low-passing, forward-backward filter.

        Args:
        - sampling_freq (int): the sampling frequency, in t_units
        - window (int): inverse of the maximum frequency (in frames).
        """
        if window == 1:
            return
        if window == 2:
            max_freq = sampling_freq / (window + 0.0000001)
        else:
            max_freq = sampling_freq / window
        coeff_1, coeff_0 = scipy.signal.iirfilter(
            4, Wn=max_freq, fs=sampling_freq, btype="low", ftype="butter"
        )
        self.matrix = np.apply_along_axis(
            lambda x: scipy.signal.filtfilt(coeff_1, coeff_0, x),
            axis=1,
            arr=self.matrix,
        )
        self.num_of_steps = self.matrix.shape[1]
        self.range = np.array([np.min(self.matrix), np.max(self.matrix)])

    def smooth_mov_av(self, window: int):
        """
        Smooths the data using a moving average with a specified window size.

        Args:
        - window (int): Size of the moving average window.
        """
        weights = np.ones(window) / window
        self.matrix = np.apply_along_axis(
            lambda x: np.convolve(x, weights, mode="valid"),
            axis=1,
            arr=self.matrix,
        )
        self.num_of_steps = self.matrix.shape[1]
        self.range = np.array([np.min(self.matrix), np.max(self.matrix)])

    def normalize(self):
        """Normalizes the data between 0 and 1."""

        data_min, data_max = self.range[0], self.range[1]
        self.matrix = (self.matrix - data_min) / (data_max - data_min)
        self.range = [np.min(self.matrix), np.max(self.matrix)]

    def create_copy(self):
        """
        Returns an independent copy of the UniData object.
        Changes to the copy will not affect the original object.
        """
        copy_data = copy.deepcopy(self)
        return copy_data

    def plot_medoids(self):
        """
        Computes and plots the average time sequence inside each environment.

        Notes:
        - Computes cluster means and standard deviations for each cluster.
        - Plots the average time sequence and standard
            deviation for each cluster.
        - Saves the figure as a PNG file in the 'output_figures' directory.
        """
        tau_window = int(self.num_of_steps / self.labels.shape[1])
        all_the_labels = self.labels
        # Initialize lists to store cluster means and standard deviations
        center_list = []
        std_list = []
        env0 = []

        # If there are no assigned window, we still need the "0" state
        # for consistency:
        list_of_labels = np.unique(all_the_labels)
        if 0 not in list_of_labels:
            list_of_labels = np.insert(list_of_labels, 0, 0)

        # Loop through unique labels (clusters)
        for ref_label in list_of_labels:
            tmp = []
            # Iterate through molecules and their labels
            for i, mol in enumerate(all_the_labels):
                for window, label in enumerate(mol):
                    # Define time interval
                    time_0 = window * tau_window
                    time_1 = (window + 1) * tau_window
                    # If the label matches the current cluster,
                    # append the corresponding data to tmp
                    if label == ref_label:
                        tmp.append(self.matrix[i][time_0:time_1])

            # Calculate mean and standard deviation for the current cluster
            if len(tmp) > 0 and ref_label > 0:
                center_list.append(np.mean(tmp, axis=0))
                std_list.append(np.std(tmp, axis=0))
            elif len(tmp) > 0:
                env0 = tmp

        center_arr = np.array(center_list)
        std_arr = np.array(std_list)
        np.savetxt(
            "medoid_center.txt",
            center_arr,
            header="Signal average for each ENV",
        )
        np.savetxt(
            "medoid_stddev.txt",
            std_arr,
            header="Signal standard deviation for each ENV",
        )

        # Create a color palette
        palette = []
        cmap = plt.get_cmap(COLORMAP, list_of_labels.size)
        palette.append(rgb2hex(cmap(0)))
        for i in range(1, cmap.N):
            rgba = cmap(i)
            palette.append(rgb2hex(rgba))

        # Plot
        fig, axes = plt.subplots()
        time_seq = range(tau_window)
        for center_id, center in enumerate(center_list):
            err_inf = center - std_list[center_id]
            err_sup = center + std_list[center_id]
            axes.fill_between(
                time_seq,
                err_inf,
                err_sup,
                alpha=0.25,
                color=palette[center_id + 1],
            )
            axes.plot(
                time_seq,
                center,
                label=f"ENV{center_id + 1}",
                marker="o",
                c=palette[center_id + 1],
                zorder=1,
            )

        for window in env0:
            axes.plot(
                time_seq,
                window,
                lw=0.1,
                c=palette[0],
                zorder=0,
                alpha=0.25,
            )

        fig.suptitle("Average time sequence inside each environments")
        axes.set_xlabel(r"Time $t$ [frames]")
        axes.set_ylabel(r"Signal")
        axes.xaxis.set_major_locator(MaxNLocator(integer=True))
        axes.legend(loc="lower left")

        fig.savefig("output_figures/Fig4.png", dpi=600)


class MultiData:
    """
    The input signals of the analysis.
    """

    def __init__(self, filename: str):
        if filename.endswith((".npz", ".npy")):
            try:
                if filename.endswith(".npz"):
                    with np.load(filename) as data:
                        data_name = data.files[0]
                        data_array = data[data_name]
                else:
                    data_array = np.load(filename)
                print("\tOriginal data shape:", data_array.shape)
            except OSError as exc_msg:
                print(f"\tReading data from {filename}: {exc_msg}")
                return
            except ValueError as exc_msg:
                print(f"\tReading data from {filename}: {exc_msg}")
                return
        else:
            print("\tERROR: unsupported format for input file.")
            return

        for dim, _ in enumerate(data_array[:-1]):
            if data_array[dim].shape != data_array[dim + 1].shape:
                print("ERROR: The signals do not correspond. Abort.")
                return

        self.dims = data_array.shape[0]
        self.num_of_particles = data_array.shape[1]
        self.num_of_steps = data_array.shape[2]
        self.range = np.array(
            [[np.min(data), np.max(data)] for data in data_array]
        )

        self.matrix: np.ndarray = np.transpose(data_array, axes=(1, 2, 0))
        self.labels = np.array([])

    def print_info(self):
        """
        Prints information about the input data.
        """
        print("Number of particles:", self.num_of_particles)
        print("Number of steps:", self.num_of_steps)
        print("Number of components:", self.dims)
        print("Data range:", self.range)

    def remove_delay(self, t_delay: int):
        """
        Removes a specified time delay from the data.

        Args:
        - t_delay (int): Number of steps to remove from the beginning.
        """
        self.matrix = self.matrix[:, t_delay:, :]
        self.num_of_steps = self.matrix.shape[1]

    def smooth(self, window: int):
        """
        Smooths the data using a moving average with a specified window size.

        Args:
        - window (int): Size of the moving average window.
        """
        weights = np.ones(window) / window
        tmp_matrix = np.transpose(self.matrix, axes=(2, 0, 1))
        tmp_matrix = np.apply_along_axis(
            lambda x: np.convolve(x, weights, mode="valid"),
            axis=2,
            arr=tmp_matrix,
        )
        self.matrix = np.transpose(tmp_matrix, axes=(1, 2, 0))
        self.num_of_steps = self.matrix.shape[1]
        self.range = np.array(
            [[np.min(comp), np.max(comp)] for comp in tmp_matrix]
        )

    def normalize(self, dim_to_avoid: list[int]):
        """Normalizes the data between 0 and 1."""
        tmp_matrix = np.transpose(self.matrix, axes=(2, 0, 1))
        new_matrix = []
        for dim, comp in enumerate(tmp_matrix):
            if dim not in dim_to_avoid:
                data_min, data_max = np.min(comp), np.max(comp)
                new_matrix.append((comp - data_min) / (data_max - data_min))
            else:
                new_matrix.append(comp)
        self.matrix = np.transpose(np.array(new_matrix), axes=(1, 2, 0))
        self.range = np.array(
            [[np.min(comp), np.max(comp)] for comp in new_matrix]
        )

    def create_copy(self):
        """
        Returns an independent copy of the UniData object.
        Changes to the copy will not affect the original object.
        """
        copy_data = copy.deepcopy(self)
        return copy_data

    def plot_medoids(self):
        """
        Plot the mean time sequence for clusters in the data.
        """
        if self.dims > 2:
            print("plot_medoids() does not work with 3D data.")
            return

        # Initialize lists to store cluster means and standard deviations
        tau_window = int(self.num_of_steps / self.labels.shape[1])
        all_the_labels = self.labels
        center_list = []
        env0 = []

        # If there are no assigned window, we still need the "0" state
        # for consistency:
        list_of_labels = np.unique(all_the_labels)
        if 0 not in list_of_labels:
            list_of_labels = np.insert(list_of_labels, 0, 0)

        # Loop through unique labels (clusters)
        for ref_label in np.unique(all_the_labels):
            tmp = []
            # Iterate through molecules and their labels
            for i, mol in enumerate(all_the_labels):
                for window, label in enumerate(mol):
                    # Define time interval
                    time_0 = window * tau_window
                    time_1 = (window + 1) * tau_window
                    # If the label matches the current cluster,
                    # append the corresponding data to tmp
                    if label == ref_label:
                        tmp.append(self.matrix[i][time_0:time_1])

            # Calculate mean for the current cluster
            if len(tmp) > 0 and ref_label > 0:
                center_list.append(np.mean(tmp, axis=0))
            elif len(tmp) > 0:
                env0 = tmp

        center_arr = np.array(center_list)
        np.save(
            "medoid_center.npy",
            center_arr,
        )

        # Create a color palette
        palette = []
        cmap = plt.get_cmap(COLORMAP, list_of_labels.size)
        palette.append(rgb2hex(cmap(0)))
        for i in range(1, cmap.N):
            rgba = cmap(i)
            palette.append(rgb2hex(rgba))

        # Plot
        fig, axes = plt.subplots()
        for id_c, center in enumerate(center_list):
            sig_x = center[:, 0]
            sig_y = center[:, 1]
            axes.plot(
                sig_x,
                sig_y,
                label=f"ENV{id_c + 1}",
                marker="o",
                c=palette[id_c + 1],
            )
        for window in env0:
            axes.plot(
                window.T[0],
                window.T[1],
                lw=0.1,
                c=palette[0],
                zorder=0,
                alpha=0.25,
            )

        fig.suptitle("Average time sequence inside each environments")
        axes.set_xlabel(r"Signal 1")
        axes.set_ylabel(r"Signal 2")
        axes.xaxis.set_major_locator(MaxNLocator(integer=True))
        axes.legend()

        fig.savefig("output_figures/Fig4.png", dpi=600)


class Parameters:
    """
    Contains the set of parameters for the specific analysis.
    """

    def __init__(self, input_file: str):
        try:
            with open(input_file, "r", encoding="utf-8") as file:
                lines = file.readlines()
        except OSError as exc_msg:
            print(f"\tReading input_parameters.txt: {exc_msg}")
        except ValueError as exc_msg:
            print(f"\tReading input_parameters.txt: {exc_msg}")

        ### Ste the default values ###
        self.t_smooth = 1
        self.t_delay = 0
        self.t_conv = 1.0
        self.t_units = "[frames]"
        self.example_id = 0
        self.bins: Union[str, int] = "auto"
        self.num_tau_w = 20
        self.min_tau_w = 2
        self.max_tau_w = -1
        self.min_t_smooth = 1
        self.max_t_smooth = 5
        self.step_t_smooth = 1

        for line in lines:
            key, value = [s for s in line.strip().split("\t") if s != ""]
            if key == "tau_window":
                self.tau_w = int(value)
            elif key == "t_smooth":
                self.t_smooth = int(value)
            elif key == "t_delay":
                self.t_delay = int(value)
            elif key == "t_conv":
                self.t_conv = float(value)
            elif key == "t_units":
                self.t_units = r"[" + str(value) + r"]"
            elif key == "example_ID":
                self.example_id = int(value)
            elif key == "bins":
                self.bins = int(value)
            elif key == "num_tau_w":
                self.num_tau_w = int(value)
            elif key == "min_tau_w":
                self.min_tau_w = int(value)
            elif key == "max_tau_w":
                self.max_tau_w = int(value)
            elif key == "min_t_smooth":
                self.min_t_smooth = int(value)
            elif key == "max_t_smooth":
                self.max_t_smooth = int(value)
            elif key == "step_t_smooth":
                self.step_t_smooth = int(value)

    def print_time(self, num_of_steps: int):
        """
        Generates time values based on parameters and number of steps.

        Args:
        - num_of_steps (int): Number of time steps.

        Returns:
        - np.ndarray: Array of time values.
        """
        t_start = self.t_delay + int(self.t_smooth / 2)
        time = (
            np.linspace(t_start, t_start + num_of_steps, num_of_steps)
            * self.t_conv
        )
        return time

    def print_to_screen(self):
        """
        Prints to screen all the analysis' parameters.
        """
        print("\n########################")
        print("### Input parameters ###")
        print("# tau_window = ", self.tau_w)
        print("# t_smooth = ", self.t_smooth)
        print("# t_delay = ", self.t_delay)
        print("# t_conv = ", self.t_conv)
        print("# t_units = ", self.t_units)
        print("# example_ID = ", self.example_id)
        print("# bins = ", self.bins)
        print("# num_tau_w = ", self.num_tau_w)
        print("# min_tau_w = ", self.min_tau_w)
        if self.max_tau_w > 1:
            print("# max_tau_w = ", self.max_tau_w)
        print("# min_t_smooth = ", self.min_t_smooth)
        print("# max_t_smooth = ", self.max_t_smooth)
        print("# step_t_smooth = ", self.step_t_smooth)
        print("########################\n")

    def create_copy(self):
        """
        Returns an independent copy of the Parameter object.
        Changes to the copy will not affect the original object.
        """
        copy_par = copy.deepcopy(self)
        return copy_par



"""
Example script for running onion_clustering
"""

import os
import shutil
from pathlib import Path

from onion_clustering import main

##############################################################################
### Set all the analysis parameters ###
# Use git clone git@github.com:matteobecchi/onion_example_files.git
# to download example datasets
PATH_TO_INPUT_DATA = "../onion_example_files/data/univariate_time-series.npy"
TAU_WINDOW = 4  # time resolution of the analysis

### Optional parametrers ###
T_SMOOTH = 1  # window for moving average (default 1)
T_DELAY = 1  # remove the first t_delay frames (default 0)
T_CONV = 1.0  # convert frames in time units (default 1)
TIME_UNITS = "frames"  # the time units (default 'frames')
EXAMPLE_ID = 0  # particle plotted as example (default 0)
NUM_TAU_W = 20  # number of values of tau_window tested (default 20)
MIN_TAU_W = 2  # min number of tau_window tested (default 2)
MIN_T_SMOOTH = 1  # min value of t_smooth tested (default 1)
MAX_T_SMOOTH = 2  # max value of t_smooth tested (default 5)
STEP_T_SMOOTH = 1  # increment in value of t_smooth tested (default 1)
MAX_TAU_W = "auto"  # max number of tau_window tested (default is auto)
BINS = "auto"  # number of histogram bins (default is auto)
##############################################################################

### Create the output directory and move there ###
original_dir = Path.cwd()
output_path = Path("./onion_output")
if output_path.exists():
    shutil.rmtree(output_path)
output_path.mkdir()
os.chdir(output_path)

try:
    ### Create the 'data_directory.txt' file ###
    with open("data_directory.txt", "w+", encoding="utf-8") as file:
        print(PATH_TO_INPUT_DATA, file=file)

    ### Create the 'input_parameter.txt' file ###
    with open("input_parameters.txt", "w+", encoding="utf-8") as file:
        print("tau_window\t" + str(TAU_WINDOW), file=file)
        print("t_smooth\t" + str(T_SMOOTH), file=file)
        print("t_delay\t" + str(T_DELAY), file=file)
        print("t_conv\t" + str(T_CONV), file=file)
        print("t_units\t" + TIME_UNITS, file=file)
        print("example_ID\t" + str(EXAMPLE_ID), file=file)
        print("num_tau_w\t" + str(NUM_TAU_W), file=file)
        print("min_tau_w\t" + str(MIN_TAU_W), file=file)
        print("min_t_smooth\t" + str(MIN_T_SMOOTH), file=file)
        print("max_t_smooth\t" + str(MAX_T_SMOOTH), file=file)
        print("step_t_smooth\t" + str(STEP_T_SMOOTH), file=file)
        if MAX_TAU_W != "auto":
            print("max_tau_w\t" + str(MAX_TAU_W), file=file)
        if BINS != "auto":
            print("bins\t" + str(BINS), file=file)

    ### Perform the clustering analysis ###
    cl_ob = main.main()

    ### Plot the output figures in output_figures/ ###

    # Plots number of states and fraction_0 as a function of tau_window
    cl_ob.plot_tra_figure()

    # Plots the raw data
    cl_ob.plot_input_data("Fig0")

    # Plots the data with the clustering thresholds and Gaussians
    cl_ob.plot_cumulative_figure()

    # Plots the colored signal for the particle with `example_ID` ID
    cl_ob.plot_one_trajectory()

    # Plots the mean time sequence inside each state
    cl_ob.data.plot_medoids()

    # Plots the population of each state as a function of time
    cl_ob.plot_state_populations()

    # Plots the Sankey diagram between the input time_windows
    cl_ob.sankey([0, 10, 20, 30, 40])

    # Writes the files for the visualization of the colored trj
    if os.path.exists("../trajectory.xyz"):
        cl_ob.print_colored_trj_from_xyz("../trajectory.xyz")
    else:
        cl_ob.print_labels()
finally:
    os.chdir(original_dir)



"""
Should contains all the functions in common between the 2 codes.
"""

from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
import scipy.signal
from scipy.integrate import quad
from scipy.optimize import OptimizeWarning

from onion_clustering.first_classes import Parameters, StateMulti, StateUni


def read_input_data() -> str:
    """
    Attempt to read the content of 'data_directory.txt' file
    and load it into a string.

    Returns
    -------

    data_dir : str
        The path to the input file.
    """
    try:
        data_dir = np.loadtxt("data_directory.txt", dtype=str)
    except OSError as msg_exc:
        print(f"\t Reading data_directory.txt: {msg_exc}")
    except ValueError as msg_exc:
        print(f"\t Reading data_directory.txt: {msg_exc}")

    print("* Reading data from", data_dir)

    return str(data_dir)


def moving_average(
    data: np.ndarray,
    window: int,
) -> np.ndarray:
    """
    Applies a moving average filter to a 1D or 2D NumPy array.

    Parameters
    ----------

    data : np.ndarray
        The input array to be smoothed.

    window : int
        The size of the moving average window.

    Returns
    -------

    np.ndarray
        The smoothed array obtained after applying
        the moving average filter.
    """
    weights = np.ones(window) / window
    if data.ndim == 1:
        return np.convolve(data, weights, mode="valid")
    if data.ndim >= 2:
        return np.apply_along_axis(
            lambda x: np.convolve(x, weights, mode="valid"), axis=1, arr=data
        )
    raise ValueError(
        "Invalid array dimension. Only 1D and 2D arrays are supported."
    )


def moving_average_2d(
    data: np.ndarray,
    side: int,
) -> np.ndarray:
    """Applies a 2D moving average filter to a NumPy array.

    Parameters
    ----------

    data : np.ndarray
        The 2D input array to be smoothed.

    side : int
        The side length of the square moving average window
        (must be an odd number).

    Returns
    -------

    np.ndarray
        The smoothed array obtained after applying the 2D
        moving average filter.
    """
    if side % 2 == 0:  # Check if side is an odd number
        raise ValueError("L must be an odd number.")
    half_width = (side - 1) // 2
    result = np.zeros_like(data, dtype=float)

    for index in np.ndindex(*data.shape):
        slices = tuple(
            slice(
                max(0, i - half_width),
                min(data.shape[dim], i + half_width + 1),
            )
            for dim, i in enumerate(index)
        )
        subarray = data[slices]
        # Calculate the average if the subarray is not empty
        if subarray.size > 0:
            result[index] = subarray.mean()

    return result


def plot_histo(
    axes: plt.Axes,
    counts: np.ndarray,
    bins: np.ndarray,
):
    """Plots a histogram on the specified axes.

    Parameters
    ----------

    axes
        The matplotlib axes to plot on.

    counts : np.ndarray
        The count or frequency of occurrences.

    bins : np.ndarray
        The bin edges defining the intervals.
    """
    axes.stairs(counts, bins, fill=True)
    axes.set_xlabel(r"Normalized signal")
    axes.set_ylabel(r"Probability distribution")


def param_grid(
    par: Parameters,
    trj_len: int,
) -> Tuple[List, List]:
    """Generates parameter grids for tau_window and t_smooth.

    Parameters
    ----------

    par : Parameters
        An instance of the Parameters class containing parameter details.

    trj_len : int
        Length of the trajectory data.

    Returns
    -------

    tau_window : List[int]
        A list of tau_window values.

    t_smooth : List[int]
        A list of t_smooth values.
    """
    if par.max_tau_w == -1:
        par.max_tau_w = trj_len - par.max_t_smooth
    tmp = np.geomspace(
        par.min_tau_w, par.max_tau_w, num=par.num_tau_w, dtype=int
    )
    tau_window = []
    for tau_w in tmp:
        if tau_w not in tau_window:
            tau_window.append(tau_w)
    print("* Tau_w used:", tau_window)

    t_smooth = list(
        range(par.min_t_smooth, par.max_t_smooth + 1, par.step_t_smooth)
    )
    print("* t_smooth used:", t_smooth)

    return tau_window, t_smooth


def gaussian(
    x_points: np.ndarray, x_mean: float, sigma: float, area: float
) -> np.ndarray:
    """Compute the Gaussian function values at given points 'x'.

    Parameters
    ----------

    x_points : np.ndarray
        Array of input values.

    x_mean : float
        Mean value of the Gaussian function.

    sigma : float
        Standard deviation of the Gaussian function.

    area : float
        Area under the Gaussian curve.

    Returns
    -------

    np.ndarray
        Gaussian function values computed at the input points.
    """
    return (
        np.exp(-(((x_points - x_mean) / sigma) ** 2))
        * area
        / (np.sqrt(np.pi) * sigma)
    )


def find_minima_around_max(
    data: np.ndarray, max_ind: Tuple[int, ...], gap: int
):
    """
    Find minima surrounding the maximum value in the given data array.

    Args:
    - data (np.ndarray): Input data array.
    - max_ind (tuple): Indices of the maximum value in the data.
    - gap (int): Gap value to determine the search range
        around the maximum.

    Returns:
    - list: List of indices representing the minima surrounding
        the maximum in each dimension.
    """
    minima: List[int] = []

    for dim in range(data.ndim):
        min_id0 = max(max_ind[dim] - gap, 0)
        min_id1 = min(max_ind[dim] + gap, data.shape[dim] - 1)

        tmp_max1: List[int] = list(max_ind)
        tmp_max2: List[int] = list(max_ind)

        tmp_max1[dim] = min_id0
        tmp_max2[dim] = min_id0 - 1
        while min_id0 > 0 and data[tuple(tmp_max1)] > data[tuple(tmp_max2)]:
            tmp_max1[dim] -= 1
            tmp_max2[dim] -= 1
            min_id0 -= 1

        tmp_max1 = list(max_ind)
        tmp_max2 = list(max_ind)

        tmp_max1[dim] = min_id1
        tmp_max2[dim] = min_id1 + 1
        while (
            min_id1 < data.shape[dim] - 1
            and data[tuple(tmp_max1)] > data[tuple(tmp_max2)]
        ):
            tmp_max1[dim] += 1
            tmp_max2[dim] += 1
            min_id1 += 1

        minima.extend([min_id0, min_id1])

    return minima


def find_half_height_around_max(
    data: np.ndarray, max_ind: Tuple[int, ...], gap: int
):
    """
    Find half-heigth points surrounding the maximum value
        in the given data array.

    Args:
    - data (np.ndarray): Input data array.
    - max_ind (tuple): Indices of the maximum value in the data.
    - gap (int): Gap value to determine the search range
        around the maximum.

    Returns:
    - list: List of indices representing the minima surrounding
        the maximum in each dimension.
    """
    max_val = data.max()
    minima: List[int] = []

    for dim in range(data.ndim):
        half_id0 = max(max_ind[dim] - gap, 0)
        half_id1 = min(max_ind[dim] + gap, data.shape[dim] - 1)

        tmp_max: List[int] = list(max_ind)

        tmp_max[dim] = half_id0
        while half_id0 > 0 and data[tuple(tmp_max)] > max_val / 2:
            tmp_max[dim] -= 1
            half_id0 -= 1

        tmp_max = list(max_ind)

        tmp_max[dim] = half_id1
        while (
            half_id1 < data.shape[dim] - 1
            and data[tuple(tmp_max)] > max_val / 2
        ):
            tmp_max[dim] += 1
            half_id1 += 1

        minima.extend([half_id0, half_id1])

    return minima


def custom_fit(
    dim: int,
    max_ind: int,
    minima: list[int],
    edges: np.ndarray,
    counts: np.ndarray,
    m_limits: np.ndarray,
) -> Tuple[int, float, np.ndarray]:
    """Fit a Gaussian curve to selected data based on provided parameters.

    Parameters
    ----------

    dim : int
        The dimension of the data.

    max_ind : int
        Index of the maximum value in the histogram.

    minima : list[int]
        List of indices representing the minimum points.

    edges : np.ndarray
        Array containing the bin edges of the histogram.

    counts : np.ndarray
        Array containing histogram counts.

    m_limits : list[list[int]]
        List of min and max limits for each dimension.

    Returns
    -------

    flag : int
        Flag indicating the success (1) or failure (0) of the fitting process.

    coeff_det_r2 : float
        Determination coefficient of the fit (r^2). Between 0 and 1.

    popt : list[float]
        Optimal values for the parameters (mu, sigma, area) of the
        fitted Gaussian.
    """
    # Extract relevant data within the specified minima
    edges_selection = edges[minima[2 * dim] : minima[2 * dim + 1]]
    all_axes = tuple(i for i in range(counts.ndim) if i != dim)
    counts_selection = np.sum(counts, axis=all_axes)
    counts_selection = counts_selection[minima[2 * dim] : minima[2 * dim + 1]]

    # Initial parameter guesses
    mu0 = edges[max_ind]
    sigma0 = (edges[minima[2 * dim + 1]] - edges[minima[2 * dim]]) / 2
    area0 = max(counts_selection) * np.sqrt(np.pi) * sigma0

    flag = 1
    coeff_det_r2 = 0
    try:
        # Attempt to fit a Gaussian using curve_fit
        popt, _, infodict, _, _ = scipy.optimize.curve_fit(
            gaussian,
            edges_selection,
            counts_selection,
            p0=[mu0, sigma0, area0],
            bounds=(
                [m_limits[dim][0], 0.0, 0.0],
                [m_limits[dim][1], np.inf, np.inf],
            ),
            full_output=True,
        )

        ss_res = np.sum(infodict["fvec"] ** 2)
        ss_tot = np.sum((counts_selection - np.mean(counts_selection)) ** 2)
        coeff_det_r2 = 1 - ss_res / ss_tot
    except OptimizeWarning:
        print("Fit: Optimize warning. ")
        flag = 0
        popt = np.empty((3,))
    except RuntimeError:
        print("\tFit: Runtime error. ")
        flag = 0
        popt = np.empty((3,))
    except TypeError:
        print("\tFit: TypeError.")
        flag = 0
        popt = np.empty((3,))
    except ValueError:
        print("\tFit: ValueError.")
        flag = 0
        popt = np.empty((3,))
    return flag, coeff_det_r2, popt


def relabel_states(
    all_the_labels: np.ndarray,
    states_list: list[StateUni],
) -> Tuple[np.ndarray, list[StateUni]]:
    """
    Relabel states and update the state list based on occurrence in
    'all_the_labels'.

    Parameters
    ----------

    all_the_labels : np.ndarray
        Array containing labels assigned to each window in the trajectory.

    states_list : list[StateUni]
        List of StateUni objects representing different states.

    Returns
    -------

    all_the_labels : np.ndarray
        Updated labels array with relabeled states.

    relevant_states : list[StateUni]
        Updated list of non-empty states, ordered by mean values.
    """
    # Step 1: Remove states with zero relevance
    relevant_states = [state for state in states_list if state.perc != 0.0]

    # Step 2: Sort states according to their mean value
    relevant_states.sort(key=lambda x: x.mean)

    # Step 3: Create a dictionary to map old state labels to new ones
    state_mapping = {
        state_index: index + 1
        for index, state_index in enumerate(
            [states_list.index(state) for state in relevant_states]
        )
    }

    relabel_map = np.zeros(len(states_list) + 1, dtype=int)
    for key, value in state_mapping.items():
        # Increment key by 1 to account for zero-indexed relabeling
        relabel_map[key + 1] = value

    # Step 4: Relabel the data in all_the_labels according to the
    # new states_list
    mask = all_the_labels != 0  # Create a mask for non-zero elements
    all_the_labels[mask] = relabel_map[all_the_labels[mask]]

    return all_the_labels, relevant_states


def find_intersection(st_0: StateUni, st_1: StateUni) -> Tuple[float, int]:
    """
    Finds the intersection between two Gaussians.

    Parameters
    ----------

    st_0, st_1 : StateUni
        The two states we are computing the threshold between.

    Returns
    -------

    th_val : float
        The value of the threshold.

    th_type : int
        The type of the threshold (1 or 2).

    Notes
    -----

    If the intersection exists, the threshold is type 1. If there are 2
    intersections, the one with higher value is chosen.
    If no intersection exists, the threshold is type 2. Its value will be
    the average between the two means, weighted with the two sigmas.
    """
    coeff_a = st_1.sigma**2 - st_0.sigma**2
    coeff_b = -2 * (st_0.mean * st_1.sigma**2 - st_1.mean * st_0.sigma**2)
    tmp_c = np.log(st_0.area * st_1.sigma / st_1.area / st_0.sigma)
    coeff_c = (
        (st_0.mean * st_1.sigma) ** 2
        - (st_1.mean * st_0.sigma) ** 2
        - ((st_0.sigma * st_1.sigma) ** 2) * tmp_c
    )
    delta = coeff_b**2 - 4 * coeff_a * coeff_c
    if coeff_a == 0.0:
        only_th = (st_0.mean + st_1.mean) / 2 - st_0.sigma**2 / 2 / (
            st_1.mean - st_0.mean
        ) * np.log(st_0.area / st_1.area)
        return only_th, 1
    if delta >= 0:
        th_plus = (-coeff_b + np.sqrt(delta)) / (2 * coeff_a)
        th_minus = (-coeff_b - np.sqrt(delta)) / (2 * coeff_a)
        # if st_0.mean < th_plus and st_1.mean > th_plus:
        #     return th_plus, 1
        intercept_plus = gaussian(th_plus, st_0.mean, st_0.sigma, st_0.area)
        intercept_minus = gaussian(th_minus, st_0.mean, st_0.sigma, st_0.area)
        if intercept_plus >= intercept_minus:
            return th_plus, 1
        return th_minus, 1
    th_aver = (st_0.mean / st_0.sigma + st_1.mean / st_1.sigma) / (
        1 / st_0.sigma + 1 / st_1.sigma
    )
    return th_aver, 2


def shared_area_between_gaussians(
    area1, mean1, sigma1, area2, mean2, sigma2
) -> Tuple[float, float]:
    """
    Computes the shared area between two Gaussians.

    Parameters
    ----------

    area1, mean1, sigma1 : float
        The parameters of Gaussian 1.

    area2, mean2, sigma2 : float
        The parameters of Gaussian 2.

    Returns
    -------

    shared_fraction_1 : float
        The fraction of the area of the first Gaussian in common with the
        second Gaussian.

    shared_fraction_2 : float
        The fraction of the area of the second Gaussian in common with the
        first Gaussian.
    """

    def gauss_1(x):
        gauss = gaussian(x, mean1, sigma1, area1)
        return gauss

    def gauss_2(x):
        gauss = gaussian(x, mean2, sigma2, area2)
        return gauss

    def min_of_gaussians(x):
        min_values = np.minimum(
            gaussian(x, mean1, sigma1, area1),
            gaussian(x, mean2, sigma2, area2),
        )
        return min_values

    area_gaussian_1, _ = quad(gauss_1, -np.inf, np.inf)
    area_gaussian_2, _ = quad(gauss_2, -np.inf, np.inf)
    shared_area, _ = quad(min_of_gaussians, -np.inf, np.inf)
    shared_fraction_1 = shared_area / area_gaussian_1
    shared_fraction_2 = shared_area / area_gaussian_2

    return shared_fraction_1, shared_fraction_2


def final_state_settings(
    list_of_states: List[StateUni],
    m_range: np.ndarray,
) -> List[StateUni]:
    """
    Final adjustemts and output in the list of identified states.

    Parameters
    ----------

    list_of_states : list[StateUni]
        The list of final states.

    m_range : np.ndarray of shape (2,)
        Range of values in the data matrix.

    Returns
    -------

    list_of_states : list[StateUni]
        Now with the correct thresholds asssigned to each state.
    """
    # Calculate the final threshold values
    # and their types based on the intercept between neighboring states.
    list_of_states[0].th_inf[0] = m_range[0]
    list_of_states[0].th_inf[1] = 0

    for i in range(len(list_of_states) - 1):
        th_val, th_type = find_intersection(
            list_of_states[i], list_of_states[i + 1]
        )
        list_of_states[i].th_sup[0] = th_val
        list_of_states[i].th_sup[1] = th_type
        list_of_states[i + 1].th_inf[0] = th_val
        list_of_states[i + 1].th_inf[1] = th_type

    list_of_states[-1].th_sup[0] = m_range[1]
    list_of_states[-1].th_sup[1] = 0

    # Write the final states and final thresholds to text files.
    with open("final_states.txt", "a", encoding="utf-8") as file:
        print("####################################", file=file)
        print("# Mu \t Sigma \t A \t state_fraction", file=file)
        for state in list_of_states:
            print(state.mean, state.sigma, state.area, state.perc, file=file)
    with open("final_thresholds.txt", "a", encoding="utf-8") as file:
        print("####################################", file=file)
        print("# Threshold_value \t Threshold type", file=file)
        for state in list_of_states:
            print(state.th_inf[0], state.th_inf[1], file=file)
        print(
            list_of_states[-1].th_sup[0],
            list_of_states[-1].th_sup[1],
            file=file,
        )

    return list_of_states


def set_final_states(
    list_of_states: List[StateUni],
    all_the_labels: np.ndarray,
    area_max_overlap: float,
) -> Tuple[List[StateUni], np.ndarray]:
    """
    Assigns final states and relabels labels based on specific criteria.

    Parameters
    ----------

    list_of_states : List[StateUni]
        List of StateUni objects representing potential states.

    all_the_labels : np.ndarray of shape (n_particles, n_windows)
        The proposed labels for each data point.

    area_max_overlap : float
        The threshold for merging two states together if they overlap.

    Returns
    -------

    updated_states : List[StateUni]
        Final list of StateUni objects representing potential states.

    all_the_labels : np.ndarray of shape (n_particles, n_windows)
        The definitive labels for each data point.
    """
    # Find all the possible merges: j could be merged into i --> [j, i]
    proposed_merge = []
    for i, st_0 in enumerate(list_of_states):
        for j, st_1 in enumerate(list_of_states):
            if j > i:
                # Condition 1: area overlap
                shared_area_1, shared_area_2 = shared_area_between_gaussians(
                    st_1.area,
                    st_1.mean,
                    st_1.sigma,
                    st_0.area,
                    st_0.mean,
                    st_0.sigma,
                )
                thresh = area_max_overlap
                if shared_area_1 > thresh >= shared_area_2:
                    proposed_merge.append([j, i])
                elif shared_area_2 > thresh >= shared_area_1:
                    proposed_merge.append([i, j])
                elif shared_area_1 > thresh and shared_area_2 > thresh:
                    proposed_merge.append(
                        [j, i] if shared_area_1 > shared_area_2 else [i, j]
                    )
                # Condition 2: mean proximity
                elif (
                    st_0.peak > st_1.peak
                    and np.abs(st_0.mean - st_1.mean) < st_0.sigma
                    and st_1.sigma < 2 * st_0.sigma
                ):
                    proposed_merge.append([j, i])
                elif (
                    st_1.peak > st_0.peak
                    and np.abs(st_0.mean - st_1.mean) < st_1.sigma
                    and st_0.sigma < 2 * st_1.sigma
                ):
                    proposed_merge.append([i, j])

    # Find the best merges (merge into the closest candidate)
    best_merge = []
    states_to_be_merged = np.unique([pair[0] for pair in proposed_merge])
    for j in states_to_be_merged:
        candidate_merge = []
        for pair in proposed_merge:
            if pair[0] == j:
                candidate_merge.append(pair)
        if len(candidate_merge) == 1:
            best_merge.append(candidate_merge[0])
        else:
            importance = [
                list_of_states[pair[1]].perc for pair in candidate_merge
            ]
            best_merge.append(candidate_merge[np.argmax(importance)])

    # Settle merging chains
    # if [i, j], all the [k, i] become [k, j]
    for pair in best_merge:
        for j, elem in enumerate(best_merge):
            if elem[1] == pair[0] and elem[0] != pair[1]:
                best_merge[j][1] = pair[1]

    # Relabel the labels in all_the_labels
    relabel_dic = {}
    for pair in best_merge:
        relabel_dic[pair[0]] = pair[1]
    if_env0 = np.any(np.unique(all_the_labels) == 0)

    relabel_map = np.zeros(max(np.unique(all_the_labels) + 1), dtype=int)
    for i, _ in enumerate(relabel_map):
        relabel_map[i] = i
    for key, value in relabel_dic.items():
        relabel_map[key + 1] = value + 1

    all_the_labels = relabel_map[all_the_labels.flatten()].reshape(
        all_the_labels.shape
    )

    final_map = np.zeros(max(np.unique(all_the_labels)) + 1, dtype=int)
    for i, elem in enumerate(np.unique(all_the_labels)):
        final_map[elem] = i + 1 * (1 - if_env0)
    for i, particle in enumerate(all_the_labels):
        for j, elem in enumerate(particle):
            all_the_labels[i][j] = final_map[elem]

    # Remove merged states from the state list
    states_to_remove = set(s0 for s0, s1 in best_merge)
    updated_states = [
        state
        for i, state in enumerate(list_of_states)
        if i not in states_to_remove
    ]

    return updated_states, all_the_labels


def find_max_prob_state(
    window: np.ndarray,
    old_label: int,
    list_of_states: List[StateUni],
) -> int:
    """
    Assign a single window to the state for which the belonging
    is the most probable.

    Parameters
    ----------

    window : np.ndarray of shape (tau_window,)
        The signal window to assign to a state.

    old_label : int
        The temporary label for the considered signal window.

    list_of_states : List[StateUni]
        List of the identified states.

    Returns
    -------

    new_label : int
        The label for the considered signal window.

    Notes
    -----

    I am using the meadian here, instead of the mean, because it's more
    robust against outliers. Not sure if this is the best chioce.
    """
    median_x = np.median(window)
    new_label = old_label
    state = list_of_states[old_label - 1]
    gauss_max = gaussian(median_x, state.mean, state.sigma, state.area)
    for i, state in enumerate(list_of_states):
        gauss = gaussian(median_x, state.mean, state.sigma, state.area)
        if gauss > gauss_max:
            new_label = i + 1
    return new_label


def max_prob_assignment(
    list_of_states: List[StateUni],
    matrix: np.ndarray,
    all_the_labels: np.ndarray,
    m_range: np.ndarray,
    tau_window: int,
    number_of_sigmas: float,
) -> Tuple[np.ndarray, List[StateUni]]:
    """
    After all the states have been identified, assign each window.
    Each signal window is assigned to the most probable state.

    Parameters
    ----------

    list_of_states : List[StateUni]
        List of the identified states.

    matrix : np.ndarray of shape (num_of_particles, num_of_timesteps)
        The data to cluster.

    all_the_labels : np.ndarray of shape (num_of_particles, num_of_windows)
        The temporary labels assigned to the signal windows.

    m_range : np.ndarray of shape (2,)
        Range of values in the data matrix.

    tau_window : int
        The time resolution of the analysis.

    Returns
    -------

    final_labels : np.ndarray of shape (num_of_particles, num_of_windows)
        The definitive labels for all the signal windows.

    updated_states : List[StateUni]
        List of the identified states, with updated percetages.
    """
    final_labels = np.zeros(all_the_labels.shape, dtype=int)
    for i, mol in enumerate(all_the_labels):
        for j, old_label in enumerate(mol):
            if old_label > 0:
                window = matrix[i][tau_window * j : tau_window * (j + 1)]
                new_label = find_max_prob_state(
                    window, old_label, list_of_states
                )
                s_range = (
                    2.0
                    * number_of_sigmas
                    * list_of_states[new_label - 1].sigma
                )
                if np.max(window) - np.min(window) < s_range:
                    final_labels[i][j] = new_label

    for i, state in enumerate(list_of_states):
        num_of_points = np.sum(final_labels == i + 1)
        state.perc = num_of_points / final_labels.size

    states_to_remove = []
    for i, state in enumerate(list_of_states):
        if state.perc == 0.0:
            states_to_remove.append(i)

    for i in states_to_remove[::-1]:
        list_of_states.pop(i)

    updated_states = final_state_settings(
        list_of_states,
        m_range,
    )

    return final_labels, updated_states


def relabel_states_2d(
    all_the_labels: np.ndarray, states_list: list[StateMulti]
) -> Tuple[np.ndarray, List[StateMulti]]:
    """
    Reorders labels and merges strongly overlapping states in a
    multidimensional space.

    Args:
    - all_the_labels (np.ndarray): An ndarray containing labels associated
        with each state.
    - states_list (list[StateMulti]): A list of StateMulti objects
        representing states to be evaluated.

    Returns:
    - tuple[np.ndarray, list[StateMulti]]: A tuple containing:
        - Updated ndarray of labels reflecting the changes in state indices.
        - Modified list of StateMulti objects after merging and relabeling
        states.
    """

    ### Step 1: Remove states with zero relevance
    sorted_states = [state for state in states_list if state.perc != 0.0]

    ### Step 2: Sort states according to their relevance
    # Create a dictionary to map old state labels to new ones
    state_mapping = {
        index: i + 1
        for i, index in enumerate(
            np.argsort([-state.perc for state in sorted_states])
        )
    }

    sorted_states.sort(key=lambda x: x.perc, reverse=True)

    # Relabel the data in all_the_labels according to the new states_list
    mask = all_the_labels != 0  # Create a mask for non-zero elements
    all_the_labels[mask] = np.vectorize(state_mapping.get, otypes=[int])(
        all_the_labels[mask] - 1, 0
    )

    ### Step 3: Merge together the states which are strongly overlapping
    # Find all the possible merges
    proposed_merge = []
    for i, st_0 in enumerate(sorted_states):
        for j, st_1 in enumerate(sorted_states):
            if j > i:
                diff = np.abs(np.subtract(st_1.mean, st_0.mean))
                if np.all(
                    diff
                    < [
                        max(st_0.sigma[k], st_1.sigma[k])
                        for k in range(diff.size)
                    ]
                ):
                    proposed_merge.append([j, i])

    # Find the best merges (merge into the most important candidate)
    best_merge = []
    states_to_be_merged = np.unique([pair[0] for pair in proposed_merge])
    for j in states_to_be_merged:
        candidate_merge = []
        for pair in proposed_merge:
            if pair[0] == j:
                candidate_merge.append(pair)
        if len(candidate_merge) == 1:
            best_merge.append(candidate_merge[0])
        else:
            # list_of_distances = [
            #     np.linalg.norm(
            #         sorted_states[pair[1]].mean - sorted_states[pair[0]].mean
            #     )
            #     for pair in candidate_merge
            # ]
            # best_merge.append(candidate_merge[np.argmin(list_of_distances)])
            importance = [
                sorted_states[pair[1]].perc for pair in candidate_merge
            ]
            best_merge.append(candidate_merge[np.argmax(importance)])

    # Settle merging chains
    # if [i, j], all the [k, i] become [k, j]
    for pair in best_merge:
        for j, elem in enumerate(best_merge):
            if elem[1] == pair[0] and elem[0] != pair[1]:
                best_merge[j][1] = pair[1]

    # Relabel the labels in all_the_labels
    relabel_dic = {}
    for pair in best_merge:
        relabel_dic[pair[0]] = pair[1]
    relabel_map = np.zeros(max(np.unique(all_the_labels) + 1), dtype=int)
    for i, _ in enumerate(relabel_map):
        relabel_map[i] = i
    for key, value in relabel_dic.items():
        relabel_map[key + 1] = value + 1

    # Remove the gaps in the labeling
    tmp_labels = np.unique(relabel_map)
    map2 = np.zeros(max(tmp_labels + 1), dtype=int)
    for i, elem in enumerate(tmp_labels):
        map2[elem] = i
    for i, elem in enumerate(relabel_map):
        relabel_map[i] = map2[elem]

    all_the_labels = relabel_map[all_the_labels.flatten()].reshape(
        all_the_labels.shape
    )

    # Remove merged states from the state list
    states_to_remove = set(s0 for s0, s1 in best_merge)
    updated_states = [
        state
        for i, state in enumerate(sorted_states)
        if i not in states_to_remove
    ]

    # Compute the fraction of data points in each state
    for st_id, state in enumerate(updated_states):
        num_of_points = np.sum(all_the_labels == st_id + 1)
        state.perc = num_of_points / all_the_labels.size

    ### Step 4: print informations on the final states
    with open("final_states.txt", "w", encoding="utf-8") as file:
        print("#center_coords, semiaxis, fraction_of_data", file=file)
        for state in updated_states:
            centers = "[" + str(state.mean[0]) + ", "
            for tmp in state.mean[1:-1]:
                centers += str(tmp) + ", "
            centers += str(state.mean[-1]) + "]"
            axis = "[" + str(state.axis[0]) + ", "
            for tmp in state.axis[1:-1]:
                axis += str(tmp) + ", "
            axis += str(state.axis[-1]) + "]"
            print(centers, axis, state.perc, file=file)

    return all_the_labels, updated_states



"""
Code for clustering of univariate time-series data.
See the documentation for all the details.
"""

import copy
import os
import shutil
import warnings
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal
from scipy.optimize import OptimizeWarning

from onion_clustering.classes import ClusteringObject1D
from onion_clustering.first_classes import Parameters, StateUni, UniData
from onion_clustering.functions import (
    gaussian,
    max_prob_assignment,
    moving_average,
    param_grid,
    plot_histo,
    read_input_data,
    relabel_states,
    set_final_states,
)

OUTPUT_FILE = "states_output.txt"
AREA_MAX_OVERLAP = 0.8


def all_the_input_stuff(number_of_sigmas: float) -> ClusteringObject1D:
    """
    Reads input parameters and raw data from specified files and directories,
    processes the raw data, and creates output files.

    Parameters
    ----------

    number_of_sigmas : float
        The signal windos are classified inside a state with a certain mean
        and std_dev if all the points differ from the mean less than
        std_dev * number_of_sigmas.

    Returns
    -------

    clustering_object : ClusteringObject1D
        The object containing all the information and data for the analysis.
    """
    # Read input parameters from files.
    data_directory = read_input_data()
    par = Parameters("input_parameters.txt")
    par.print_to_screen()

    # Read raw data from the specified directory/files.
    if isinstance(data_directory, str):
        data = UniData(data_directory)
    else:
        print("\tERROR: data_directory.txt is missing or wrongly formatted. ")

    # Remove initial frames based on 't_delay'.
    data.remove_delay(par.t_delay)

    ### Create files for output
    with open(OUTPUT_FILE, "w", encoding="utf-8") as file:
        print("#", file=file)
    figures_folder = "output_figures"
    if not os.path.exists(figures_folder):
        os.makedirs(figures_folder)
    for filename in os.listdir(figures_folder):
        file_path = os.path.join(figures_folder, filename)
        try:
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except OSError as ex_msg:
            print(f"Failed to delete {file_path}. Reason: {ex_msg}")

    clustering_object = ClusteringObject1D(par, data, number_of_sigmas)

    return clustering_object


def perform_gauss_fit(
    param: List[int],
    data: List[np.ndarray],
    int_type: str,
) -> Tuple[bool, float, np.ndarray]:
    """
    Perform Gaussian fit on given data.

    Parameters
    ----------

    id0 : int
        Index representing the lower limit for data selection.

    id1 : int
        Index representing the upper limit for data selection.

    bins : np.ndarray
        Array containing bin values.

    counts : np.ndarray
        Array containing counts corresponding to bins.

    n_data : int
        Number of data points.

    interval_type : str
        Type of interval ("min" of "half").

    Returns
    -------

    bool
        True if the fit is successful, False otherwise.

    coeff_det_r2 : float
        Coefficient of determination of the fit. Zero if the fit fails.

    popt : np.ndarray or None
        Parameters of the Gaussian fit if successful, None otherwise.
    """
    id0, id1, max_ind, n_data = param
    bins, counts = data

    selected_bins = bins[id0:id1]
    selected_counts = counts[id0:id1]
    mu0 = bins[max_ind]
    sigma0 = (bins[id1] - bins[id0]) / 6
    area0 = counts[max_ind] * np.sqrt(np.pi) * sigma0
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            popt, _, infodict, _, _ = scipy.optimize.curve_fit(
                gaussian,
                selected_bins,
                selected_counts,
                p0=[mu0, sigma0, area0],
                full_output=True,
            )
            if popt[1] < 0:
                popt[1] *= -1
                popt[2] *= -1
            popt[2] *= n_data
            ss_res = np.sum(infodict["fvec"] ** 2)
            ss_tot = np.sum((selected_counts - np.mean(selected_counts)) ** 2)
            coeff_det_r2 = 1 - ss_res / ss_tot
            return True, coeff_det_r2, popt
    except OptimizeWarning:
        print(f"\t{int_type} fit: Optimize warning. ")
        return (
            False,
            0,
            np.empty(
                3,
            ),
        )
    except RuntimeError:
        print(f"\t{int_type} fit: Runtime error. ")
        return (
            False,
            0,
            np.empty(
                3,
            ),
        )
    except TypeError:
        print(f"\t{int_type} fit: TypeError.")
        return (
            False,
            0,
            np.empty(
                3,
            ),
        )
    except ValueError:
        print(f"\t{int_type} fit: ValueError.")
        return (
            False,
            0,
            np.empty(
                3,
            ),
        )


def gauss_fit_max(
    m_clean: np.ndarray,
    par: Parameters,
    number_of_sigmas: float,
    filename: str,
    full_out: bool,
) -> Union[StateUni, None]:
    """
    Performs Gaussian fitting on input data.

    Parameters
    ----------

    m_clean : np.ndarray
        Input data for Gaussian fitting.

    par : Parameters
        Object containing parameters for fitting.

    number_of_sigmas : float
        To set the thresholds for assigning windows to the state.

    filename : str
        Name of the output plot file.

    full_output : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    Returns
    -------

    state : StateUni
        Object containing Gaussian fit parameters (mu, sigma, area)
        or None if the fit fails.
    """
    print("* Gaussian fit...")
    flat_m = m_clean.flatten()

    ### 1. Histogram ###
    counts, bins = np.histogram(flat_m, bins=par.bins, density=True)
    gap = 1
    if bins.size > 49:
        gap = int(bins.size * 0.02) * 2
    print(f"\tNumber of bins = {bins.size}, gap = {gap}")

    ### 2. Smoothing with tau = 3 ###
    counts = moving_average(counts, gap)
    bins = moving_average(bins, gap)
    if (counts == 0.0).any():
        print(
            "\tWARNING: there are empty bins. "
            "Consider reducing the number of bins."
        )

    ### 3. Find the maximum ###
    max_val = counts.max()
    max_ind = counts.argmax()

    ### 4. Find the minima surrounding it ###
    min_id0 = np.max([max_ind - gap, 0])
    min_id1 = np.min([max_ind + gap, counts.size - 1])
    while min_id0 > 0 and counts[min_id0] > counts[min_id0 - 1]:
        min_id0 -= 1
    while min_id1 < counts.size - 1 and counts[min_id1] > counts[min_id1 + 1]:
        min_id1 += 1

    ### 5. Try the fit between the minima and check its goodness ###
    fit_param = [min_id0, min_id1, max_ind, flat_m.size]
    fit_data = [bins, counts]
    flag_min, r_2_min, popt_min = perform_gauss_fit(fit_param, fit_data, "Min")

    ### 6. Find the inrterval of half height ###
    half_id0 = np.max([max_ind - gap, 0])
    half_id1 = np.min([max_ind + gap, counts.size - 1])
    while half_id0 > 0 and counts[half_id0] > max_val / 2:
        half_id0 -= 1
    while half_id1 < counts.size - 1 and counts[half_id1] > max_val / 2:
        half_id1 += 1

    ### 7. Try the fit between the minima and check its goodness ###
    fit_param = [half_id0, half_id1, max_ind, flat_m.size]
    fit_data = [bins, counts]
    flag_half, r_2_half, popt_half = perform_gauss_fit(
        fit_param, fit_data, "Half"
    )

    ### 8. Choose the best fit ###
    r_2 = r_2_min
    if flag_min == 1 and flag_half == 0:
        popt = popt_min
    elif flag_min == 0 and flag_half == 1:
        popt = popt_half
        r_2 = r_2_half
    elif flag_min * flag_half == 1:
        if r_2_min >= r_2_half:
            popt = popt_min
        else:
            popt = popt_half
            r_2 = r_2_half
    else:
        print("\tWARNING: this fit is not converging.")
        return None

    state = StateUni(popt[0], popt[1], popt[2])
    state.build_boundaries(number_of_sigmas)

    with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
        print("\n", file=file)
        print(
            f"\tmu = {state.mean:.4f}, sigma = {state.sigma:.4f},"
            f" area = {state.area:.4f}"
        )
        print(
            f"\tmu = {state.mean:.4f}, sigma = {state.sigma:.4f},"
            f" area = {state.area:.4f}",
            file=file,
        )
        print(f"\tFit r2 = {r_2}", file=file)

    if full_out:
        ### Plot the distribution and the fitted gaussians
        y_spread = np.max(m_clean) - np.min(m_clean)
        y_lim = [
            np.min(m_clean) - 0.025 * y_spread,
            np.max(m_clean) + 0.025 * y_spread,
        ]
        fig, axes = plt.subplots()
        plot_histo(axes, counts, bins)
        axes.set_xlim(y_lim)
        tmp_popt = [state.mean, state.sigma, state.area / flat_m.size]
        axes.plot(
            np.linspace(bins[0], bins[-1], 1000),
            gaussian(np.linspace(bins[0], bins[-1], 1000), *tmp_popt),
        )

        fig.savefig(filename + ".png", dpi=600)
        plt.close(fig)

    return state


def find_stable_trj(
    cl_ob: ClusteringObject1D,
    state: StateUni,
    tmp_labels: np.ndarray,
    lim: int,
) -> Tuple[np.ndarray, float, bool]:
    """
    Identifies stable windows within a state.

    Parameters
    ----------

    cl_ob : ClusteringObject1D

    state : StateUni
        Object containing stable state parameters.

    all_the_labels : np.ndarray
        Labels indicating window classifications.

    offset : int
        Offset value for classifying stable windows.

    Returns
    -------

    m2_array : np.ndarray
        Array of non-stable windows.

    fw : float
        Fraction of windows classified as stable.

    one_last_state : bool
        Indicates if there's one last state remaining.
    """
    print("* Finding stable windows...")

    # Calculate the number of windows in the trajectory
    number_of_windows = tmp_labels.shape[1]
    m_clean = cl_ob.data.matrix
    tau_window = cl_ob.par.tau_w

    mask_unclassified = tmp_labels < 0.5
    m_reshaped = m_clean[:, : number_of_windows * tau_window].reshape(
        m_clean.shape[0], number_of_windows, tau_window
    )
    mask_inf = np.min(m_reshaped, axis=2) >= state.th_inf[0]
    mask_sup = np.max(m_reshaped, axis=2) <= state.th_sup[0]
    mask = mask_unclassified & mask_inf & mask_sup

    tmp_labels[mask] = lim + 1
    counter = np.sum(mask)

    # Initialize an empty list to store non-stable windows
    remaning_data = []
    mask_remaining = mask_unclassified & ~mask
    for i, window in np.argwhere(mask_remaining):
        r_w = m_clean[i, window * tau_window : (window + 1) * tau_window]
        remaning_data.append(r_w)

    # Calculate the fraction of stable windows found
    window_fraction = counter / (tmp_labels.size)

    # Print the fraction of stable windows
    with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
        print(
            f"\tFraction of windows in state {lim + 1}"
            f" = {window_fraction:.3}"
        )
        print(
            f"\tFraction of windows in state {lim + 1}"
            f" = {window_fraction:.3}",
            file=file,
        )

    # Convert the list of non-stable windows to a NumPy array
    m2_array = np.array(remaning_data)
    one_last_state = True
    if m2_array.size == 0:
        one_last_state = False

    # Return the array of non-stable windows, the fraction of stable windows,
    # and the updated list_of_states
    return m2_array, window_fraction, one_last_state


def solve_batman(
    cl_ob: ClusteringObject1D,
    m_clean: np.ndarray,
    par: Parameters,
    number_of_sigmas: float,
    tmp_labels: np.ndarray,
    tau_window: int,
    lim: int,
):
    """
    This functions takes care of particular cases where the
    data points on the tails of a Gaussian are not correctly assigned,
    creating weird sharp peaks in the hsistogram.

    Parameters
    ----------

    cl_ob : ClusteringObject1D

    m_clean : np.ndarray
        Input data for Gaussian fitting.

    par : Parameters
        Object containing parameters for fitting.

    number_of_sigmas : float
        To set the thresholds for assigning windows to the state.

    tmp_labels : np.ndarray
        Labels indicating window classifications.

    tau_windw : int
        The time resolution of the analysis.

    lim : int
        Offset value for classifying stable windows.
    """
    flat_m = m_clean.flatten()

    ### 1. Histogram ###
    counts, bins = np.histogram(flat_m, bins=par.bins, density=True)
    gap = 1
    if bins.size > 99:
        gap = int(bins.size * 0.02)
    print(f"\tNumber of bins = {bins.size}, gap = {gap}")

    ### 2. Smoothing with tau = 3 ###
    counts = moving_average(counts, gap)
    bins = moving_average(bins, gap)
    if (counts == 0.0).any():
        print(
            "\tWARNING: there are empty bins. "
            "Consider reducing the number of bins."
        )

    ### 3. Find the maxima ###
    max_ind, _ = scipy.signal.find_peaks(counts)
    max_val = np.array([counts[i] for i in max_ind])

    for i, m_ind in enumerate(max_ind[:1]):
        ### 4. Find the minima surrounding it ###
        min_id0 = np.max([m_ind - gap, 0])
        min_id1 = np.min([m_ind + gap, counts.size - 1])
        while min_id0 > 0 and counts[min_id0] > counts[min_id0 - 1]:
            min_id0 -= 1
        while (
            min_id1 < counts.size - 1 and counts[min_id1] > counts[min_id1 + 1]
        ):
            min_id1 += 1

        ### 5. Try the fit between the minima and check its goodness ###
        fit_param = [min_id0, min_id1, m_ind, flat_m.size]
        fit_data = [bins, counts]
        flag_min, r_2_min, popt_min = perform_gauss_fit(
            fit_param, fit_data, "Min"
        )

        ### 6. Find the inrterval of half height ###
        half_id0 = np.max([m_ind - gap, 0])
        half_id1 = np.min([m_ind + gap, counts.size - 1])
        while half_id0 > 0 and counts[half_id0] > max_val[i] / 2:
            half_id0 -= 1
        while half_id1 < counts.size - 1 and counts[half_id1] > max_val[i] / 2:
            half_id1 += 1

        ### 7. Try the fit between the minima and check its goodness ###
        fit_param = [half_id0, half_id1, m_ind, flat_m.size]
        fit_data = [bins, counts]
        flag_half, r_2_half, popt_half = perform_gauss_fit(
            fit_param, fit_data, "Half"
        )

        ### 8. Choose the best fit ###
        if flag_min == 1 and flag_half == 0:
            popt = popt_min
        elif flag_min == 0 and flag_half == 1:
            popt = popt_half
        elif flag_min * flag_half == 1:
            if r_2_min >= r_2_half:
                popt = popt_min
            else:
                popt = popt_half
        else:
            continue

        state = StateUni(popt[0], popt[1], popt[2])
        state.build_boundaries(number_of_sigmas)

        # Calculate the number of windows in the trajectory
        number_of_windows = tmp_labels.shape[1]

        mask_unclassified = tmp_labels < 0.5
        m_reshaped = cl_ob.data.matrix[
            :, : number_of_windows * tau_window
        ].reshape(cl_ob.data.matrix.shape[0], number_of_windows, tau_window)
        mask_inf = np.min(m_reshaped, axis=2) >= state.th_inf[0]
        mask_sup = np.max(m_reshaped, axis=2) <= state.th_sup[0]
        mask = mask_unclassified & mask_inf & mask_sup

        tmp_labels[mask] = lim + 1
        counter = np.sum(mask)

        # Initialize an empty list to store non-stable windows
        remaning_data = []
        mask_remaining = mask_unclassified & ~mask
        for i, window in np.argwhere(mask_remaining):
            r_w = cl_ob.data.matrix[
                i, window * tau_window : (window + 1) * tau_window
            ]
            remaning_data.append(r_w)

        # Calculate the fraction of stable windows found
        window_fraction = counter / (tmp_labels.size)

        # Print the fraction of stable windows
        with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
            print(
                f"\tFraction of windows in state {lim + 1}"
                f" = {window_fraction:.3}"
            )
            print(
                f"\tFraction of windows in state {lim + 1}"
                f" = {window_fraction:.3}",
                file=file,
            )

        # Convert the list of non-stable windows to a NumPy array
        m2_array = np.array(remaning_data)
        one_last_state = True
        if len(m2_array) == 0:
            one_last_state = False

        # Return the array of non-stable windows, the fraction of stable windows,
        # and the updated list_of_states
        return state, m2_array, window_fraction, one_last_state

    # If event trying all the maxima does not work, surrend
    return None, None, None, None


def iterative_search(
    cl_ob: ClusteringObject1D,
    name: str,
    full_out: bool,
) -> Tuple[ClusteringObject1D, np.ndarray, bool]:
    """
    Performs an iterative search for stable states in a trajectory.

    Parameters
    ----------

    cl_ob : ClusteringObject1D

    name : str
        Name for identifying output figures.

    full_output : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    Returns
    -------

    cl_ob : ClusteringObject1D
        Updated with the clustering results.

    atl : np.ndarray of shape (n_particles, n_windows)
        Temporary array of labels.

    one_last_state : bool
        Indicates if there's one last state remaining.
    """

    # Initialize an array to store labels for each window.
    num_windows = int(cl_ob.data.num_of_steps / cl_ob.par.tau_w)
    tmp_labels = np.zeros((cl_ob.data.num_of_particles, num_windows)).astype(
        int
    )

    with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
        print(f"tau_window = {cl_ob.par.tau_w}", file=file)

    states_list = []
    m_copy = cl_ob.data.matrix
    iteration_id = 1
    states_counter = 0
    one_last_state = False
    while True:
        ### Locate and fit maximum in the signal distribution
        state = gauss_fit_max(
            m_copy,
            cl_ob.par,
            cl_ob.number_of_sigmas,
            f"output_figures/{name}Fig1_{iteration_id}",
            full_out,
        )
        if state is None:
            print("Iterations interrupted because fit does not converge. ")
            break

        ### Find the windows in which the trajectories are stable
        m_next, counter, one_last_state = find_stable_trj(
            cl_ob, state, tmp_labels, states_counter
        )

        ### Exit the loop if no new stable windows are found
        if counter <= 0.0:
            state, m_next, counter, one_last_state = solve_batman(
                cl_ob,
                m_copy,
                cl_ob.par,
                cl_ob.number_of_sigmas,
                tmp_labels,
                cl_ob.par.tau_w,
                states_counter,
            )

            if counter == 0 or state is None:
                print("Iterations interrupted because last state is empty. ")
                break

        state.perc = counter
        states_list.append(state)
        states_counter += 1
        iteration_id += 1
        m_copy = m_next

        if m_next.size == 0:
            print("Iterations interrupted because all points are classififed.")
            break

    cl_ob.iterations = len(states_list)
    atl, lis = relabel_states(tmp_labels, states_list)
    cl_ob.states = lis
    return cl_ob, atl, one_last_state


def timeseries_analysis(
    cl_ob: ClusteringObject1D,
    tau_w: int,
    t_smooth: int,
    full_out: bool,
) -> Tuple[int, float, List[float]]:
    """
    Performs an analysis pipeline on time series data.

    Parameters
    ----------

    cl_ob : ClusteringObject1D

    tau_w : int
        The time resolution for the analysis.

    t_smooth : int
        The width of the moving average for the analysis.

    Returns
    -------

    num_states : int
        Number of identified states.

    fraction_0 : float
        Fraction of unclassified data points. Between 0 and 1.

    list_of_pop : List[float]
        List of the populations of the different states.
    """
    print("* New analysis: ", tau_w, t_smooth)
    name = str(t_smooth) + "_" + str(tau_w) + "_"

    tmp_cl_ob = copy.deepcopy(cl_ob)
    tmp_cl_ob.par.tau_w = tau_w
    tmp_cl_ob.par.t_smooth = t_smooth

    tmp_cl_ob.preparing_the_data()
    tmp_cl_ob.plot_input_data(name + "Fig0")

    tmp_cl_ob, tmp_labels, _ = iterative_search(tmp_cl_ob, name, full_out)

    if len(tmp_cl_ob.states) == 0:
        print("* No possible classification was found. ")
        # We need to free the memory otherwise it accumulates
        del tmp_cl_ob
        return 0, 1.0, [1.0]

    list_of_states, tmp_labels = set_final_states(
        tmp_cl_ob.states,
        tmp_labels,
        AREA_MAX_OVERLAP,
    )

    tmp_cl_ob.data.labels, tmp_cl_ob.states = max_prob_assignment(
        list_of_states,
        tmp_cl_ob.data.matrix,
        tmp_labels,
        tmp_cl_ob.data.range,
        tau_w,
        tmp_cl_ob.number_of_sigmas,
    )

    list_of_pop = [state.perc for state in tmp_cl_ob.states]
    fraction_0 = 1 - np.sum(list_of_pop)
    list_of_pop.insert(0, fraction_0)
    n_states = len(tmp_cl_ob.states)

    # We need to free the memory otherwise it accumulates
    del tmp_cl_ob

    print(f"Number of states identified: {n_states}, [{fraction_0}]\n")
    return n_states, fraction_0, list_of_pop


def full_output_analysis(
    cl_ob: ClusteringObject1D,
    full_out: bool,
) -> ClusteringObject1D:
    """
    Perform a comprehensive analysis on the input data.

    Parameters
    ----------

    cl_ob : ClusteringObject1D

    full_out : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    Returns
    -------

    cl_ob : ClusteringObject1D
        Updated with the clustering results.
    """

    cl_ob.preparing_the_data()

    cl_ob, tmp_labels, _ = iterative_search(cl_ob, "", full_out)
    if len(cl_ob.states) == 0:
        print("* No possible classification was found. ")
        return cl_ob

    list_of_states, tmp_labels = set_final_states(
        cl_ob.states,
        tmp_labels,
        AREA_MAX_OVERLAP,
    )

    cl_ob.data.labels, cl_ob.states = max_prob_assignment(
        list_of_states,
        cl_ob.data.matrix,
        tmp_labels,
        cl_ob.data.range,
        cl_ob.par.tau_w,
        cl_ob.number_of_sigmas,
    )

    return cl_ob


def time_resolution_analysis(
    cl_ob: ClusteringObject1D,
    full_out: bool,
):
    """
    Performs Temporal Resolution Analysis (TRA) to explore parameter
    space and analyze the dataset.

    Parameters
    ----------
    cl_ob : ClusteringObject1D

    full_out : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.
    """
    tau_window_list, t_smooth_list = param_grid(
        cl_ob.par, cl_ob.data.num_of_steps
    )
    cl_ob.tau_window_list = np.array(tau_window_list)
    cl_ob.t_smooth_list = np.array(t_smooth_list)

    number_of_states = []
    fraction_0 = []
    list_of_pop: List[List[List[float]]] = [
        [[] for _ in tau_window_list] for _ in t_smooth_list
    ]

    for i, tau_w in enumerate(tau_window_list):
        tmp = [tau_w]
        tmp1 = [tau_w]
        for j, t_s in enumerate(t_smooth_list):
            n_s, f_0, l_pop = timeseries_analysis(cl_ob, tau_w, t_s, full_out)
            tmp.append(n_s)
            tmp1.append(f_0)
            list_of_pop[j][i] = l_pop
        number_of_states.append(tmp)
        fraction_0.append(tmp1)

    number_of_states_arr = np.array(number_of_states)
    fraction_0_arr = np.array(fraction_0)

    np.savetxt(
        "number_of_states.txt",
        number_of_states,
        fmt="%i",
        delimiter="\t",
        header="tau_window\t number_of_states for different t_smooth",
    )
    np.savetxt(
        "fraction_0.txt",
        fraction_0,
        delimiter=" ",
        header="tau_window\t fraction in ENV0 for different t_smooth",
    )

    cl_ob.number_of_states = number_of_states_arr
    cl_ob.fraction_0 = fraction_0_arr
    cl_ob.list_of_pop = list_of_pop


def main(
    full_output: bool = True,
    number_of_sigmas: float = 2.0,
) -> ClusteringObject1D:
    """
    Returns the clustering object with the analysi.

    Parameters
    ----------

    full_output : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    number_of_sigmas : float
        The signal windos are classified inside a state with a certain mean
        and std_dev if all the points differ from the mean less than
        std_dev * number_of_sigmas.

    Returns
    -------

    clustering_object : ClusteringObject1D
        The clusteriong object, with the input data, the parameters and the
        results of the clustering.

    Notes
    -----

    all_the_input_stuff() reads the data and the parameters

    time_resolution_analysis() explore the parameter
        (tau_window, t_smooth) space.

    full_output_analysis() performs a detailed analysis
        with the chosen parameters.
    """
    print("##############################################################")
    print("# If you publish results using onion-clustering, please cite #")
    print("# this work: https://doi.org/10.48550/arXiv.2402.07786.      #")
    print("##############################################################")

    clustering_object = all_the_input_stuff(number_of_sigmas)
    time_resolution_analysis(clustering_object, full_output)
    clustering_object = full_output_analysis(clustering_object, full_output)

    return clustering_object


if __name__ == "__main__":
    main()
